#include <papi.h>

#define CHECK_PAPI(err) check_papi(err, __LINE__)

struct counters
{
    int event_set;
    long long *values;
    long long start_time;
    double elapsed_time;
    char event_names[RAPL_MAX_COUNTERS][PAPI_MAX_STR_LEN];
    char units[RAPL_MAX_COUNTERS][PAPI_MIN_STR_LEN];
    int num_events;
};

union {
    long long ll;
    double d;
} result;

void check_papi(int err, int line) {
    if (err == PAPI_OK)
        return;
    
    char *error = PAPI_strerror(err);
    printf("Papi error on line %d: %s\n", line, error);
    free(error);
    exit(1);
}

void print_counters(struct counters *counters, int verbose) {
    if (verbose) printf("Energy measurements:\n");
    for (int i = 0; i < counters->num_events; i++) {
        if (strcmp(counters->units[i], "nJ") == 0) {
            if (verbose) printf("%s\t%.6fJ\t(Average Power %.1fW)\n", counters->event_names[i], (double)counters->values[i]/1.0e9, ((double)counters->values[i]/1.0e9)/counters->elapsed_time);
            else printf("%.6f ", (double) counters->values[i]/1.0e9);
        }
    }
    if (verbose) printf("\nFixed values:\n");
    for (int i = 0; i < counters->num_events; i++) {
        if (strcmp(counters->units[i], "nJ") != 0) {
            result.ll = counters->values[i];
            if (verbose) printf("%s\t%0.3lf%s\n", counters->event_names[i], result.d, counters->units[i]);
            else printf("%0.3lf ", result.d);
        }
    }
}

void start_listening(struct counters *counters) {
    int ret;
    counters->event_set = PAPI_NULL;
    int event_code = PAPI_NATIVE_MASK;
    const PAPI_component_info_t *comp_info = NULL;
    PAPI_event_info_t event_info;
    
    ret = PAPI_library_init(PAPI_VER_CURRENT);
    int num_comps = PAPI_num_components();
    
    int rapl_comp_id;
    for (rapl_comp_id = 0; rapl_comp_id < num_comps; rapl_comp_id++) {
        comp_info = PAPI_get_component_info(rapl_comp_id);
        if (comp_info == NULL) {
            printf("Get component info failed\n");
            exit(1);
        }
        
        if (strcmp(comp_info->name, "rapl") == 0) {
            if (comp_info->num_native_events == 0) {
                printf("No rapl events found\n");
                exit(1);
            }
            break;
        }
    }
    
    if (rapl_comp_id == num_comps) {
        printf("No rapl component found\n");
        exit(1);
    }
    
    ret = PAPI_create_eventset(&counters->event_set);
    CHECK_PAPI(ret);
    
    ret = PAPI_enum_cmp_event(&event_code, PAPI_ENUM_FIRST, rapl_comp_id);
    
    for (counters->num_events = 0; ret == PAPI_OK; counters->num_events++) {
        ret = PAPI_event_code_to_name(event_code, counters->event_names[counters->num_events]);
        CHECK_PAPI(ret);
        ret = PAPI_get_event_info(event_code, &event_info);
        CHECK_PAPI(ret);
        
        strncpy(counters->units[counters->num_events], event_info.units, PAPI_MIN_STR_LEN);
        ret = PAPI_add_event(counters->event_set, event_code);
        if (ret != PAPI_OK)
            break; // event limit
        
        ret = PAPI_enum_cmp_event(&event_code, PAPI_ENUM_EVENTS, rapl_comp_id);
    }
    
    counters->values = calloc(counters->num_events, sizeof(*counters->values));
    
    counters->start_time = PAPI_get_real_nsec();
    ret = PAPI_start(counters->event_set);
    CHECK_PAPI(ret);
}

void stop_listening(struct counters *counters) {
    int ret;
    long long end_time;
    
    end_time = PAPI_get_real_nsec();
    ret = PAPI_stop(counters->event_set, counters->values);
    CHECK_PAPI(ret);
    
    counters->elapsed_time = ((double) (end_time - counters->start_time))/1.0e9;
    
    ret = PAPI_cleanup_eventset(counters->event_set);
    CHECK_PAPI(ret);
    ret = PAPI_destroy_eventset(&counters->event_set);
    CHECK_PAPI(ret);
}

