#include <papi.h>
#include <signal.h>
#include <sys/time.h>

#define CHECK_PAPI(err) check_papi(err, __LINE__)
#define RAPL_MAX_COUNTERS 64
#define REFRESH_INTERVAL 5

struct counters
{
    int energy_event_set;
    long long *energy_values;
    char event_names[RAPL_MAX_COUNTERS][PAPI_MAX_STR_LEN];
    char units[RAPL_MAX_COUNTERS][PAPI_MIN_STR_LEN];
    int num_events;

    int cache_event_set;
    long long *cache_values;

    long long start_time;
    double elapsed_time;
};

static struct counters counters;

union {
    long long ll;
    double d;
} result;

void check_papi(int err, int line) {
    if (err == PAPI_OK)
        return;
    
    char *error = PAPI_strerror(err);
    printf("Papi error on line %d: %s\n", line, error);
    free(error);
    exit(1);
}

void refresh(int sig) {
    long long values[counters.num_events];
    
    PAPI_stop(counters.energy_event_set, values);
    PAPI_start(counters.energy_event_set);
    
    for (int i = 0; i < counters.num_events; i++)
        counters.energy_values[i] += values[i];
}

void print_counters(int verbose) {
    if (verbose) {
        printf("Elapsed time: %lf\n", counters.elapsed_time);
        printf("Energy measurements:\n");
    } else {
        printf("%lf ", counters.elapsed_time);
    }
    for (int i = 0; i < counters.num_events; i++) {
        if (strcmp(counters.units[i], "nJ") == 0) {
            if (verbose) printf("%s\t%.6fJ\t(Average Power %.1fW)\n", counters.event_names[i], (double)counters.energy_values[i]/1.0e9, ((double)counters.energy_values[i]/1.0e9)/counters.elapsed_time);
            else printf("%.6f ", (double) counters.energy_values[i]/1.0e9);
        }
    }
    
    if (verbose) {
        printf("\nCache counters:\n");
        printf("%s\t%lld\n", "L2 accesses", counters.cache_values[0]);
        printf("%s\t%lld\n", "L2 misses  ", counters.cache_values[1]);
        printf("%s\t%lld\n", "L3 accesses", counters.cache_values[2]);
        printf("%s\t%lld\n", "L3 misses  ", counters.cache_values[3]);
    } else {
        for (int i = 0; i < 4; i++)
            printf("%lld ", counters.cache_values[i]);
    }

    free(counters.energy_values);
    free(counters.cache_values);
}

int get_rapl_component_id() {
    int rapl_comp_id;
    const PAPI_component_info_t *comp_info = NULL;
    int num_comps = PAPI_num_components();
    
    for (rapl_comp_id = 0; rapl_comp_id < num_comps; rapl_comp_id++) {
        comp_info = PAPI_get_component_info(rapl_comp_id);
        if (comp_info == NULL) {
            printf("Get component info failed\n");
            exit(1);
        }
        
        if (strcmp(comp_info->name, "rapl") == 0) {
            if (comp_info->num_native_events == 0) {
                printf("No rapl events found\n");
                exit(1);
            }
            break;
        }
    }
    if (rapl_comp_id == num_comps) {
        printf("No rapl component found\n");
        exit(1);
    }
    return rapl_comp_id;
}

void add_rapl_events(int rapl_comp_id) {
    int event_code = PAPI_NATIVE_MASK;
    PAPI_event_info_t event_info;
    
    char name[PAPI_MAX_STR_LEN];
    char units[PAPI_MIN_STR_LEN];
    counters.num_events = 0;
    
    int ret = PAPI_enum_cmp_event(&event_code, PAPI_ENUM_FIRST, rapl_comp_id);
    
    while (ret == PAPI_OK) {
        ret = PAPI_event_code_to_name(event_code, name);
        CHECK_PAPI(ret);
        ret = PAPI_get_event_info(event_code, &event_info);
        CHECK_PAPI(ret);
        strncpy(units, event_info.units, PAPI_MIN_STR_LEN);
        
        
        if (!strcmp(units, "nJ")) {
            strncpy(counters.event_names[counters.num_events], name, PAPI_MAX_STR_LEN);
            strncpy(counters.units[counters.num_events], units, PAPI_MIN_STR_LEN);
            ret = PAPI_add_event(counters.energy_event_set, event_code);
            if (ret != PAPI_OK)
                break;
            counters.num_events++;
        }
        ret = PAPI_enum_cmp_event(&event_code, PAPI_ENUM_EVENTS, rapl_comp_id);
    }
    counters.energy_values = calloc(counters.num_events, sizeof(*counters.energy_values));
}

void add_cache_events() {
    int ret;
    ret = PAPI_add_event(counters.cache_event_set, PAPI_L2_TCA);
    CHECK_PAPI(ret);
    ret = PAPI_add_event(counters.cache_event_set, PAPI_L2_TCM);
    CHECK_PAPI(ret);
    ret = PAPI_add_event(counters.cache_event_set, PAPI_L3_TCA);
    CHECK_PAPI(ret);
    ret = PAPI_add_event(counters.cache_event_set, PAPI_L3_TCM);
    CHECK_PAPI(ret);
    
    counters.cache_values = calloc(4, sizeof(*counters.cache_values));
}

void start_listening() {
    int ret;
    counters.energy_event_set = PAPI_NULL;
    counters.cache_event_set = PAPI_NULL;
    
    ret = PAPI_library_init(PAPI_VER_CURRENT);
    int rapl_comp_id = get_rapl_component_id();
    
    ret = PAPI_create_eventset(&counters.energy_event_set);
    CHECK_PAPI(ret);
    ret = PAPI_create_eventset(&counters.cache_event_set);
    CHECK_PAPI(ret);
    
    add_rapl_events(rapl_comp_id);
    add_cache_events();
    
    counters.start_time = PAPI_get_real_nsec();
    ret = PAPI_start(counters.energy_event_set);
    CHECK_PAPI(ret);
    ret = PAPI_start(counters.cache_event_set);
    CHECK_PAPI(ret);
    
    struct itimerval tv;
    tv.it_interval.tv_usec = 0;
    tv.it_interval.tv_sec = REFRESH_INTERVAL;
    tv.it_value.tv_usec = 0;
    tv.it_value.tv_sec = REFRESH_INTERVAL;
    signal(SIGALRM, refresh);
    setitimer(0, &tv, NULL);
}

void stop_listening() {
    int ret;
    long long end_time;
    long long values[counters.num_events];
    
    
    ret = PAPI_stop(counters.energy_event_set, values);
    CHECK_PAPI(ret);
    for (int i = 0; i < counters.num_events; i++)
        counters.energy_values[i] += values[i];
    
    ret = PAPI_stop(counters.cache_event_set, counters.cache_values);
    CHECK_PAPI(ret);
    
    end_time = PAPI_get_real_nsec();
    counters.elapsed_time = ((double) (end_time - counters.start_time))/1.0e9;
    
    ret = PAPI_cleanup_eventset(counters.energy_event_set);
    CHECK_PAPI(ret);
    ret = PAPI_destroy_eventset(&counters.energy_event_set);
    CHECK_PAPI(ret);
    
    ret = PAPI_cleanup_eventset(counters.cache_event_set);
    CHECK_PAPI(ret);
    ret = PAPI_destroy_eventset(&counters.cache_event_set);
    CHECK_PAPI(ret);
}

