#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <cblas.h>

#include "../common/cmd_line.c"

float *calculate_on_device(float *a, float *b, int n, struct args_t args);

#include "../common/base.c"
#include "../common/files.c"
#ifdef PAPI
#include "../common/performance_papi.c"
#else
#include "../common/performance_no_papi.c"
#endif
#include "../common/run.c"

#pragma omp task in([block_size][block_size] a, [block_size][block_size] b) inout([block_size][block_size] c)
void matmul_task(float *a, float *b, float *c, unsigned long block_size)
{
    cblas_sgemm( CblasRowMajor, CblasNoTrans, CblasNoTrans, block_size, block_size, block_size, 1.0, a, block_size, b, block_size, 1.0, c, block_size);
}


void compute_parallel(unsigned long block_size, unsigned long number_of_blocks, float *a[number_of_blocks][number_of_blocks], float *b[number_of_blocks][number_of_blocks], float *c[number_of_blocks][number_of_blocks]) 
{ 
	unsigned long i, j, k;
    for ( i = 0; i < number_of_blocks; i++)
        for ( j = 0; j < number_of_blocks; j++)
            for ( k = 0; k < number_of_blocks; k++)
				matmul_task(a[i][k], b[k][j], c[i][j], block_size);
    
    #pragma omp taskwait
}

void convert_to_blocks(unsigned long block_size, unsigned long number_of_blocks, unsigned long n, float *linear, float *a[number_of_blocks][number_of_blocks])
{
	unsigned long i, j;
	for ( i = 0; i < n; i++)
	{
		for ( j = 0; j < n; j++)
		{
            a[i / block_size][j / block_size][(i % block_size) * block_size + j % block_size] = linear[i * n + j];
		}
	}
}

void convert_to_linear( unsigned long block_size,unsigned long number_of_blocks, unsigned long n, float *linear, float *a[number_of_blocks][number_of_blocks])
{
	unsigned long i, j;
	for ( i = 0; i < n; i++)
	{
		for ( j = 0; j < n; j++)
		{
            linear[i * n + j] = a[i / block_size][j / block_size][(i % block_size) * block_size + j % block_size];
		}
	}
}

void free_matrices( float** a, float** b, float** c, unsigned long number_of_blocks )
{
	size_t i;
	for (i = 0; i < number_of_blocks*number_of_blocks; i++)
	{
		free( a[i] );
		free( b[i] );
		free( c[i] );
	}
	
	free( a );
	free( b );
	free( c );
}

void create_blocked_input_data(int n, int block_size, int number_of_blocks, float *input_a, float *input_b, float *output_c, float ***a, float ***b, float ***c) 
{

    *a = (float**) malloc(number_of_blocks * number_of_blocks * sizeof(float*));
    *b = (float**) malloc(number_of_blocks * number_of_blocks * sizeof(float*));
    *c = (float**) malloc(number_of_blocks * number_of_blocks * sizeof(float*));
    
    for (int i = 0; i < number_of_blocks * number_of_blocks; i++) {
        (*a)[i] = (float*) malloc(block_size * block_size * sizeof(float));
        (*b)[i] = (float*) malloc(block_size * block_size * sizeof(float));
        (*c)[i] = (float*) malloc(block_size * block_size * sizeof(float));
    }
    
    convert_to_blocks(block_size, number_of_blocks, n, input_a, (void*)*a);
    convert_to_blocks(block_size, number_of_blocks, n, input_b, (void*)*b);
    convert_to_blocks(block_size, number_of_blocks, n, output_c, (void*)*c);
}



float *calculate_on_device(float *a, float *b, int n, struct args_t args) {
    float **blocked_a, **blocked_b, **blocked_c;

    unsigned long block_size = args.block_size;
    unsigned long number_of_blocks = n / block_size;
    
    float *c = calloc(n * n, sizeof(*c));
    
    create_blocked_input_data(n, block_size, number_of_blocks, a, b, c, &blocked_a, &blocked_b, &blocked_c);
    compute_parallel(block_size, number_of_blocks, (void*)blocked_a, (void*)blocked_b, (void*)blocked_c);
    convert_to_linear(block_size, number_of_blocks, n, c, (void*)blocked_c);
    free_matrices(blocked_a, blocked_b, blocked_c, number_of_blocks);
    
    return c;
}

int main(int argc, char const **argv) {
    run(argc, argv);
    return 0;
}

