#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>

#ifdef AVX
#include <immintrin.h>
#endif

#ifdef SSE
#include <xmmintrin.h>
#include <smmintrin.h>
#endif

#include "../common/cmd_line.c"

float *calculate_on_device(const float *a, float *b, int n, struct args_t args);

#include "../common/base.c"
#include "../common/files.c"
#ifdef PAPI
#include "../common/performance_papi.c"
#else
#include "../common/performance_no_papi.c"
#endif
#include "../common/run.c"


#ifdef AVX
void alloc(void **memptr, size_t size) {
    int r = posix_memalign(memptr, 32, size); // the fastest avx functions require the memory to be 32 byte aligned
    if (r != 0) exit(1);
}
#else // maybe an sse clause should be added for 16 byte alignment, but it seems malloc aligns enough anyway
void alloc(void **memptr, size_t size) {
    *memptr = malloc(size);
}
#endif


#pragma omp task
void convert_to_rbr_blocks(const float *mat, float **blocked, int n, int block_size, int number_of_blocks) {
    alloc((void**) blocked, n * n * sizeof(**blocked));
    
    int pos = 0;
	for (int block_i = 0; block_i < number_of_blocks; ++block_i)
	{
		for (int block_j = 0; block_j < number_of_blocks; ++block_j)
		{
			for (int i = 0; i < block_size; ++i)
			{
				for (int j = 0; j < block_size; ++j)
				{
					int block_pos = block_size * (block_i * n + block_j);
					int index = block_pos + i * n + j;
					(*blocked)[pos] = mat[index];
					pos++;
				}
			}
		}
	}
}

#pragma omp task
void convert_to_cbc_blocks(const float *mat, float **blocked, int n, int block_size, int number_of_blocks) {
    alloc((void**) blocked, n * n * sizeof(**blocked));
    
    int pos = 0;
	for (int block_i = 0; block_i < number_of_blocks; ++block_i)
	{
		for (int block_j = 0; block_j < number_of_blocks; ++block_j)
		{
			for (int i = 0; i < block_size; ++i)
			{
				for (int j = 0; j < block_size; ++j)
				{
					int block_pos = block_size * (block_j * n + block_i);
					int index = block_pos + j * n + i;
					(*blocked)[pos] = mat[index];
					pos++;
				}
			}
		}
	}
}

#pragma omp task
void matmul_rbr_cbc_blocks(const float *a, const float *b, float *c, int row, int col, int n, int block_size, int number_of_blocks) {
#if defined (SSE) || defined(AVX)
    float dots[8];
    int dot_mask = 0xff;
#endif

    float *c_block = &c[block_size * (row * n + col)];
    for (int block = 0; block < number_of_blocks; block++) {
        int a_block_index = block_size * block_size * number_of_blocks * row + block_size * block_size * block;
        int b_block_index = block_size * block_size * number_of_blocks * col + block_size * block_size * block;

        for (int i = 0; i < block_size; i++) {
            for (int j = 0; j < block_size; j++) {
                float dp = 0;
#ifdef SSE
                for (int k = 0; k < block_size; k += 4) {
#endif
#ifdef AVX
                for (int k = 0; k < block_size; k += 8) {
#endif
#ifdef NONE
                for (int k = 0; k < block_size; k++) {
#endif     
                    int a_index = a_block_index + i * block_size + k;
                    int b_index = b_block_index + j * block_size + k;
#ifdef SSE
                    const __m128 _a = _mm_load_ps(&a[a_index]);
                    const __m128 _b = _mm_load_ps(&b[b_index]);
                    const __m128 _d = _mm_dp_ps(_a, _b, dot_mask);
                    _mm_store_ps(dots, _d);
                    dp += dots[0];
#endif
#ifdef AVX
                    const __m256 _a = _mm256_load_ps(&a[a_index]);
                    const __m256 _b = _mm256_load_ps(&b[b_index]);
                    const __m256 _d = _mm256_dp_ps(_a, _b, dot_mask);
                    _mm256_store_ps(dots, _d);
                    dp += dots[0] + dots[4];
#endif
#ifdef NONE
                    dp += a[a_index] * b[b_index];
#endif
                }
                c_block[i*n+j] += dp;
            }
        }
    }
}



float *calculate_on_device(const float *a, float *b, int n, struct args_t args) {
    int block_size = args.block_size;
	int number_of_blocks = n / block_size;
	float *c = malloc(n * n * sizeof(*c));
    
    float *aa, *bb;
    
    convert_to_rbr_blocks(a, &aa, n, block_size, number_of_blocks);
    convert_to_cbc_blocks(b, &bb, n, block_size, number_of_blocks);
    
    #pragma omp taskwait
    
    for (int i = 0; i < number_of_blocks; i++)
        for (int j = 0; j < number_of_blocks; j++)
            matmul_rbr_cbc_blocks(aa, bb, c, i, j, n, block_size, number_of_blocks);

	#pragma omp taskwait

    free(aa);
    free(bb);
    
    return c;
}

int main(int argc, char const **argv) {
    run(argc, argv);
    return 0;
}

