#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif
#include "../common/cmd_line.c"

float *calculate_on_device(float *a, float *b, size_t n, struct args_t args);

#include "../common/base.c"
#include "../common/opencl_common.c"
#include "../common/files.c"
#ifdef PAPI
#include "../common/performance_papi.c"
#else
#include "../common/performance_no_papi.c"
#endif
#include "../common/run.c"


float *calculate_on_device(float *a, float *b, size_t n, struct args_t args) {
    cl_int err;
    cl_device_id device;
    cl_context context;
    cl_program program;
    cl_kernel transpose_kernel, mult_kernel;
    cl_command_queue queue;
    cl_mem_flags read_write, read, write;

#ifdef AVX
    cl_uint num_blocks = n/8;
#endif
#ifdef SSE
    cl_uint num_blocks = n/4;
#endif
#if defined (NONE) || defined(AUTO)
    cl_uint num_blocks = n;
#endif
    size_t transpose_size = num_blocks*(num_blocks+1)/2;
    size_t mult_global[2] = {n, n};
    
    char *build_options = malloc(200);
#ifdef NONE
    sprintf(build_options, "-cl-opt-disable -DNUMBLOCKS=%u -DSIZE=%ld", num_blocks, n);
#else
    sprintf(build_options, "-DNUMBLOCKS=%u -DSIZE=%ld", num_blocks, n);
#endif
    
    if (args.device == CL_DEVICE_TYPE_CPU) {
        read_write = CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR;
        read = CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR;
        write = CL_MEM_WRITE_ONLY;
    } else {
        read_write = CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR;
        read = CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR;
        write = CL_MEM_WRITE_ONLY;
    }
    
    device = get_device(args.device);
    context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);
    CHECKERR(err);
    queue = clCreateCommandQueue(context, device, 0, &err);
    CHECKERR(err);

#ifdef SSE
    program = build_program(&context, &device, "matmul-sse.cl", build_options, 0);
#endif

#ifdef AVX
    program = build_program(&context, &device, "matmul-avx.cl", build_options, 0);
#endif

#if defined (NONE) || defined(AUTO)
    program = build_program(&context, &device, "matmul-none.cl", build_options, 0);
#endif
    
    transpose_kernel = clCreateKernel(program, "transpose", &err);
    CHECKERR(err);
    mult_kernel = clCreateKernel(program, "mult", &err);
    CHECKERR(err);
    
    cl_mem a_buffer = clCreateBuffer(context, read, n*n * sizeof(*a), a, &err);
    cl_mem b_buffer = clCreateBuffer(context, read_write, n*n * sizeof(*b), b, &err);
    cl_mem c_buffer = clCreateBuffer(context, write, n*n * sizeof(float), NULL, &err);
    CHECKERR(err);
    
    err  = clSetKernelArg(transpose_kernel, 0, sizeof(cl_mem), &b_buffer);
    CHECKERR(err);
    
    err = clEnqueueNDRangeKernel(queue, transpose_kernel, 1, NULL, &transpose_size, NULL, 0, NULL, NULL);
    CHECKERR(err);
    
    err  = clSetKernelArg(mult_kernel, 0, sizeof(cl_mem), &a_buffer);
    err |= clSetKernelArg(mult_kernel, 1, sizeof(cl_mem), &b_buffer);
    err |= clSetKernelArg(mult_kernel, 2, sizeof(cl_mem), &c_buffer);
    CHECKERR(err);
    
    err = clEnqueueNDRangeKernel(queue, mult_kernel, 2, NULL, mult_global, NULL, 0, NULL, NULL);
    CHECKERR(err);
    
    float *result = calloc(n*n, sizeof(*result));
    err = clEnqueueReadBuffer(queue, c_buffer, CL_FALSE, 0, n*n * sizeof(*a), result, 0, NULL, NULL);
    CHECKERR(err);
    
    clFinish(queue);
    
    free(build_options);
    clReleaseMemObject(a_buffer);
    clReleaseMemObject(b_buffer);
    clReleaseMemObject(c_buffer);
    clReleaseKernel(transpose_kernel);
    clReleaseKernel(mult_kernel);
    clReleaseProgram(program);
    clReleaseCommandQueue(queue);
    clReleaseContext(context);
    
    return result;
}


int main (int argc, const char **argv) {
    run(argc, argv);
    return 0;
}
