from __future__ import print_function

import os
import sys
import time
import math
import shlex
import sqlite3
from subprocess import Popen, PIPE


def do_papi():
    uname = os.uname()
    return uname[0] == 'Linux'

base = os.getcwd()
papi = do_papi()

def parse_args():
    args = sys.argv[1:]
    imps = []
    if 'h' in args or 'a' in args: imps += 'OmpSs'
    if 'b' in args or 'a' in args: imps += 'OmpSs ATLAS'
    if 't' in args or 'a' in args: imps += 'OpenCL Transposed'
    if 's' in args or 'a' in args: imps += 'OpenCL Simple'
    
    rng = None
    t = None
    v = None
    it = 10
    for a in args:
        if a[:2] == 's=': # size as power of two, s=10
            p = int(a[2:])
            rng = (p, p+1)
        if a[:2] == 'r=': # range of sizes, r=8-12
            fr, to = map(int, a[2:].split('-'))
            rng = (fr, to+1)
        if a[:2] == 't=': # threads, t=1,4,8
            t = map(int, a[2:].split(','))
        if a[:2] == 'v=': # vectors, v=a,s,u,n
            v = a[2:].split(',')
        if a[:2] == 'i=': # iterations, i=10
            it = int(a[2:])
    
    tasks = []
    if 's' in args or 'a' in args:
        tasks += Task.open_cl_simple(v, it, rng)
    if 't' in args or 'a' in args:
        tasks += Task.open_cl_transposed(v, it, rng)
    if 'h' in args or 'a' in args:
        tasks += Task.ompss_hm(v, it, rng, t)
    if 'b' in args or 'a' in args:    
        tasks += Task.ompss_blas(v, it, rng, t)
    
    return tasks

def run(cmd, cwd=None):
    p = Popen(shlex.split(cmd), stdout=PIPE, cwd=cwd)
    return p.communicate()

def average(l):
    return sum(l)/len(l)

def median(l):
    s = sorted(l)
    return s[len(l)/2]

def std_dev(l):
    a = average(l)
    diffs = [(a-i)*(a-i) for i in l]
    return math.sqrt(average(diffs))


class Task(object):
    def __init__(self, name, sub_task, cwd, options='', exe='matmul', threads=1, sizes=(8, 11), iterations=10):
        self.name = name
        self.sub = sub_task
        self.options = options
        self.cwd = cwd
        self.exe = exe
        self.threads = threads
        self.sizes = sizes
        self.iterations = iterations
    
    def __repr__(self):
        return '<Task {0}-{1}>'.format(self.name, self.sub)
    
    def get_block_size(self, size):
        if self.name == 'OmpSs':
            if self.threads == 8:
                bs = {'256': 64,
                        '512': 64,
                        '1024': 128,
                        '2048': 256,
                        '4096': 128,
                        '8192': 256}
            if self.threads == 4:
                bs = {'256': 64,
                        '512': 128,
                        '1024': 512,
                        '2048': 256,
                        '4096': 128,
                        '8192': 256}
            if self.threads == 1:
                bs = {'256': 128,
                        '512': 256,
                        '1024': 512,
                        '2048': 512,
                        '4096': 256,
                        '8192': 256}
        if self.name == 'OmpSs-ATLAS':
            if self.threads == 8:
                return size/4
            if self.threads == 4:
                return size/2
            if self.threads == 1:
                return size
        return bs[str(size)]
    
    def run(self):
        print('{0} - {1}'.format(self.name, self.sub))
        for opt in self.options:
            os.putenv(opt[0], opt[1])
            
        for p in range(self.sizes[0], self.sizes[1]):
            size = 2**p
            block_size = self.get_block_size(size)
            cmd = './{0} -b {1} {2}'.format(self.exe, block_size, size)
            self.results = []
            run(cmd, self.cwd) # warming up
            for i in range(self.iterations):
                out, err = run(cmd, self.cwd)
                result = Result(self, size, out, err)
                result.save()
                self.results += [result]
            Result.print_result_list(size, self.results)
    
    
    @classmethod
    def opencl(cls, name, file, vectors, iterations, size_range):
        subs = {}
        if 'n' in vectors: subs['None'] = '-none'
        if 'u' in vectors: subs['Automatic'] = '-auto'
        if 's' in vectors: subs['SSE'] = '-sse'
        if 'a' in vectors: subs['AVX'] = '-avx'

        tasks = []
        for sub in subs:
            exe = 'matmul{0}'.format(subs[sub])
            tasks += [cls(name, sub, file, exe=exe, sizes=size_range, iterations=iterations)]
        return tasks

    @classmethod
    def open_cl_simple(cls, vectors, iterations, size_range):
        name = 'Simple OpenCL'
        file = 'opencl-matmul'
        return cls.opencl(name, file, vectors, iterations, size_range)
        
    @classmethod
    def open_cl_transposed(cls, vectors, iterations, size_range):
        name = 'Transposed OpenCL'
        file = 'opencl-matmul-transposed'
        return cls.opencl(name, file, vectors, iterations, size_range)
    
    
    @classmethod
    def ompss_hm(cls, vectors, iterations, size_range, threads):
        subs = {}
        if 'n' in vectors: subs['None'] = '-none'
        if 's' in vectors: subs['SSE'] = '-sse'
        if 'a' in vectors: subs['AVX'] = '-avx'
        
        opts = {}
        if 1 in threads: opts['Single thread'] = [('OMP_NUM_THREADS', '1')]
        if 4 in threads: opts['Four threads'] = [('OMP_NUM_THREADS', '4')]
        if 8 in threads: opts['Eight threads'] = [('OMP_NUM_THREADS', '8')]

        name = 'OmpSs'
        file = 'ompss-matmul'
        tasks = []
        for sub in subs:
            for opt in opts:
                exe = 'matmul' + subs[sub]
                sub_name = '{0} {1}'.format(sub, opt)
                o = opts[opt] + [('NX_DEPS', 'plain')]
                tasks += [cls(name, sub_name, file, o, exe, threads=int(o[0][1]), sizes=size_range, iterations=iterations)]
        return tasks
    
    @classmethod
    def ompss_blas(cls, vectors, iterations, size_range, threads):
        opts = {}
        if 1 in threads: opts['Single thread'] = [('OMP_NUM_THREADS', '1')]
        if 4 in threads: opts['Four threads'] = [('OMP_NUM_THREADS', '4')]
        if 8 in threads: opts['Eight threads'] = [('OMP_NUM_THREADS', '8')]
        
        name = 'OmpSs-ATLAS'
        file = 'ompss-blas'
        
        tasks = []
        for opt in opts:
            o = opts[opt] + [('NX_DEPS', 'plain')]
            tasks += [cls(name, opt, file, o, threads=int(o[0][1]), sizes=size_range, iterations=iterations)]
        return tasks
    
        


class Result(object):
    def __init__(self, task, size, out, err):
        try:
            if papi:
                if 'Warning!' in out:
                    lines = out.split('\n')
                    out = lines[-1]
                verify, time, package, pp1, pp0, l2a, l2m, l3a, l3m = out.split()
            else:
                verify, time = out.split()
                package = 0
                pp0 = 0
                pp1 = 0
        except ValueError:
            print(out)
            print(err)
            raise

        self.stdout = out
        self.size = size
        self.task = task
        self.verify = verify == '1'
        self.time = float(time)
        self.energy = float(package)
        self.pp1 = float(pp1)
        self.pp0 = float(pp0)
        self.l2a = long(l2a)
        self.l2m = long(l2m)
        self.l3a = long(l3a)
        self.l3m = long(l3m)
    
    @classmethod
    def print_headers(cls):
        print('{0:>14} {1:>9} {2:>9} {3:>9}'.format('Avg time', 'Med time', 'Std time', 'MFLOPS'), end='')
        if papi:
            print('{0:>14} {1:>9} {2:>9} {3:>13} {4:>12} {5:>12} {6:>12} {7:>12}'.format('MFLOPS/W', 'Avg pow', 'Med pow', 'Std pow', 'ED', 'EDD', 'L2 miss', 'L3 miss'), end='') 
        print()
    
    @classmethod
    def print_result_list(cls, size, results):
        times = [r.time for r in results]
        power = [r.energy for r in results]
        l2a = [r.l2a for r in results]
        l2m = [r.l2m for r in results]
        l3a = [r.l3a for r in results]
        l3m = [r.l3m for r in results]
    
        mflops = (2*(size**3)) / 1e6
        mflopss = mflops / median(times)
        print('{0:5}'.format(size), end='')
        print('{0:9.5} {1:9.5} {2:9.5} {3:12.8}'.format(average(times), median(times), std_dev(times), mflopss), end='')
        if papi:
            mflopsspw = mflops / median(power)
            ed = median(power) * median(times)
            edd = ed * median(times)
            l2_miss = median(l2m) / float(median(l2a))
            l3_miss = median(l3m) / float(median(l3a))
            print('{0:14.5} {1:9.5} {2:9.5} {3:12.8} {4:12.8} {5:12.8} {6:12.4} {7:12.4}'.format(mflopsspw, average(power), median(power), std_dev(power), ed, edd, l2_miss, l3_miss), end='')
        print()
    
    def __repr__(self):
        return '<{0}: time: {1}s, energy: {2}J>'.format(self.task, self.time, self.energy)
    
    def save(self):
        if not self.verify:
            raise Exception(self.stdout)
        conn = sqlite3.connect('results')
        cur = conn.cursor()
        cur.execute('insert into results (imp, sub, size, time, energy, l2a, l2m, l3a, l3m, pp0, pp1) values (?,?,?,?,?,?,?,?,?,?,?)', 
            (self.task.name, self.task.sub, self.size, self.time, self.energy, self.l2a, self.l2m, self.l3a, self.l3m, self.pp0, self.pp1))
        conn.commit()
        conn.close()
 
    
def main():
    Result.print_headers()
    tasks = parse_args()
    for task in tasks:
        try:
            task.run()
        except:
            print('exception, moving to next task')
            time.sleep(6)

if __name__ == '__main__':
    main()
